import java.text.NumberFormat;
import java.awt.*;
import javax.swing.*;
import java.awt.AWTEvent;
import java.awt.event.*;
import java.util.Enumeration;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.behaviors.vp.*;

/**
 *  KeyBehavior is a generic behavior class to take key presses and move a
 *  TransformGroup through a Java3D scene. The actions resulting from the key strokes
 *  are modified by using the Ctrl, Alt and Shift keys.
 *
 *  (version 1.2) removed modelviewer specificity, keeping ViewPlatformBehavior aspects.
 *  (version 1.1) changed modifier key behaviour, and made more specific for modelviewer
 *  (version 1.0) reconstructed class to make more generic.
 *
 * MODIFIED:
 * @version 1.2, 5 december 2001
 * @author  Thierry Graff, France
 *			<tigraff@chez.com>
 *
 * @version 1.1, 22 June 2001 djj
 * @author  Duncan J Jauncey, United Kingdom
 *			<dj@cryogen.com>
 *
 * @version 1.0, 25 September 1998 aajc
 * @author  Andrew AJ Cain, Swinburne University, Australia
 *          <acain@it.swin.edu.au>
 *
 * edited from code by:
 *   Gary S. Moss <moss@arl.mil>
 *   U. S. Army Research Laboratory
 *
 * CLASS NAME:
 *   KeyBehavior
 *
 */

public class KeyBehavior extends ViewPlatformBehavior{

  protected static final double FAST_SPEED = 2.0;
//  protected static final double NORMAL_SPEED = 1.0;
  protected static final double NORMAL_SPEED = 0.2;
  protected static final double SLOW_SPEED = 0.1;
//  protected static final double SLOW_SPEED = 0.5;
  private double speed = NORMAL_SPEED;

  private TransformGroup transformGroup;
  private Transform3D transform3D;
  private WakeupCondition keyCriterion;

  private final static double TWO_PI = (2.0 * Math.PI);
  private double rotateXAmount = Math.PI / 64.0;
  private double rotateYAmount = Math.PI / 64.0;
  private double rotateZAmount = Math.PI / 64.0;
  private double moveRate = 0.3;

  // Point to where the resetKey leads - since 1.2
  private Point3d resetPoint = new Point3d(0, 0, 0);
  // label where new coordinates are displayed after each move - since 1.2
  private static final String STR1 = "coords : (";
  private static final String STR2 = " ";
  private static final String STR3 = ") - dist : ";
  private static final String STR4 = " a.u. ";
  private JLabel coordLabel;
  private float scaleFactor = 1f;
  private static NumberFormat NF;
  static{
  	NF = NumberFormat.getInstance();
    NF.setMaximumFractionDigits(3);
    NF.setMinimumFractionDigits(3);
  };

  private int forwardKey 		= KeyEvent.VK_UP;
  private int backKey 			= KeyEvent.VK_DOWN;
  private int leftKey 			= KeyEvent.VK_LEFT;
  private int rightKey 			= KeyEvent.VK_RIGHT;
  private int resetKey 			= KeyEvent.VK_R;

  public KeyBehavior( TransformGroup tg ){
    transformGroup = tg;
    transform3D = new Transform3D();
  }

  public void initialize(){
    WakeupCriterion[] keyEvents = new WakeupCriterion[2];
    keyEvents[0] = new WakeupOnAWTEvent( KeyEvent.KEY_PRESSED );
    keyEvents[1] = new WakeupOnAWTEvent( KeyEvent.KEY_RELEASED );
    keyCriterion = new WakeupOr( keyEvents );
    wakeupOn( keyCriterion );
  }

  public void processStimulus( Enumeration criteria ){
    WakeupCriterion wakeup;
    AWTEvent[] event;
    while( criteria.hasMoreElements() ){
      wakeup = (WakeupCriterion) criteria.nextElement();
      if( !(wakeup instanceof WakeupOnAWTEvent) )
        continue;
      event = ((WakeupOnAWTEvent)wakeup).getAWTEvent();
      for( int i = 0; i < event.length; i++ ){
        if( event[i].getID() == KeyEvent.KEY_PRESSED ){
          processKeyEvent((KeyEvent)event[i]);
        }
      }
    }
    wakeupOn( keyCriterion );
    // addition 1.2 : set transformGroup's coords in coordLabel
    if (coordLabel != null){
	    transformGroup.getTransform(transform3D);
		Vector3f v = new Vector3f();
		transform3D.get(v); // vector 'v' recieves the translational component of 'transform3D'
	    //coordLabel.setText("( " + NF.format(v.x/scaleFactor) + " " + NF.format(v.y/scaleFactor) + " " + NF.format(v.z/scaleFactor) + " )");
	    coordLabel.setText(STR1 + NF.format(v.x/scaleFactor) + STR2 
        						 + NF.format(v.y/scaleFactor) + STR2 
                                 + NF.format(v.z/scaleFactor) + STR3
                                 + NF.format(v.length()/scaleFactor) + STR4);
    }
  }

  protected void processKeyEvent(KeyEvent event){
    int keycode = event.getKeyCode();
    if(event.isShiftDown()) speed = FAST_SPEED;
    else speed = NORMAL_SPEED;
    if( event.isAltDown() )
      altMove(keycode);
    else if( event.isControlDown())
      controlMove(keycode);
    else
      standardMove(keycode);
  }

  //moves forward backward or rotates left right

  private void standardMove(int keycode){
    if(keycode == forwardKey)
      moveForward();
    else if(keycode == backKey)
      moveBackward();
    else if(keycode == leftKey)
      rotLeft();
    else if(keycode == rightKey)
      rotRight();
    else if(keycode == resetKey)
      doReset();
  }

  //moves left right, rotate up down
  protected void altMove(int keycode){
    if(keycode == backKey)
      moveDown();
    else if(keycode == forwardKey)
      moveUp();
    else if(keycode == leftKey)
      moveLeft();
    else if(keycode == rightKey)
      moveRight();
  }

  //move up down, rot left right
  protected void controlMove(int keycode){
    if(keycode == forwardKey)
      rotUp();
    else if(keycode == backKey)
      rotDown();
    else if(keycode == leftKey)
      rollLeft();
    else if(keycode == rightKey)
      rollRight();
  }

  private void moveForward(){ doMove(new Vector3d(0.0,0.0, -getMovementRate())); }

  private void moveBackward(){ doMove(new Vector3d(0.0,0.0, getMovementRate())); }

  private void moveLeft(){ doMove(new Vector3d( -getMovementRate() ,0.0,0.0)); }

  private void moveRight(){ doMove(new Vector3d(getMovementRate(),0.0,0.0)); }

  private void moveUp(){ doMove(new Vector3d(0.0, getMovementRate() ,0.0)); }

  private void moveDown(){ doMove(new Vector3d(0.0, -getMovementRate() ,0.0)); }

  protected void rotRight(){ doRotateY(getRotateRightAmount()); }

  protected void rotUp(){ doRotateX(getRotateUpAmount()); }

  protected void rotLeft(){ doRotateY(getRotateLeftAmount()); }

  protected void rotDown(){ doRotateX(getRotateDownAmount()); }

  protected void rollLeft(){ doRotateZ(getRollLeftAmount()); }

  protected void rollRight(){ doRotateZ(getRollRightAmount()); }

  protected void doRotateY(double radians){
    transformGroup.getTransform(transform3D);
    Transform3D toMove = new Transform3D();
    toMove.rotY(radians);
    transform3D.mul(toMove);
    transformGroup.setTransform(transform3D);
  }

  protected void doRotateX(double radians){
    transformGroup.getTransform(transform3D);
    Transform3D toMove = new Transform3D();
    toMove.rotX(radians);
    transform3D.mul(toMove);
    transformGroup.setTransform(transform3D);
  }

  protected void doRotateZ(double radians){
    transformGroup.getTransform(transform3D);
    Transform3D toMove = new Transform3D();
    toMove.rotZ(radians);
    transform3D.mul(toMove);
    transformGroup.setTransform(transform3D);
  }

  protected void doMove(Vector3d theMove){
    transformGroup.getTransform(transform3D);
    Transform3D toMove = new Transform3D();
    toMove.setTranslation(theMove);
    transform3D.mul(toMove);
    transformGroup.setTransform(transform3D);
  }

  protected void doReset(){
	  Vector3d move = new Vector3d(resetPoint);
	  Transform3D t3d = new Transform3D();
	  //transform.set( move );
	  t3d.lookAt(resetPoint, new Point3d(0, 0, 0), new Vector3d(0, 0, 1));
	  t3d.invert();
	  transformGroup.setTransform(t3d);
  }

  protected double getMovementRate(){ return moveRate * speed; }

  protected double getRollLeftAmount(){ return rotateZAmount * speed; }

  protected double getRollRightAmount(){ return -rotateZAmount * speed; }

  protected double getRotateUpAmount(){ return rotateYAmount * speed; }

  protected double getRotateDownAmount(){ return -rotateYAmount * speed; }

  protected double getRotateLeftAmount(){ return rotateYAmount * speed; }

  protected double getRotateRightAmount(){ return -rotateYAmount * speed; }

  /** 
  Permits to set the position occupied after reseting
  @since 1.2
  */
  public void setResetPosition(Point3d p){ resetPoint = p; }

  /** 
  Permits to set the label where the coords are displayed
  @since 1.2
  */
  public void setCoordLabel(JLabel l){ coordLabel = l; }

  /** 
  Permits to set the scale factor by which the displayed coordinates must be divided.
  @since 1.2
  */
  public void setScaleFactor(float sf){ scaleFactor = sf; }

  public void setRotateXAmount(double radians){ rotateXAmount = radians; }

  public void setRotateYAmount(double radians){ rotateYAmount = radians; }

  public void setRotateZAmount(double radians){ rotateZAmount = radians; }

  public void setMovementRate(double meters){
  	moveRate = meters; // Travel rate in meters/frame
  }

  public void setForwardKey(int key){ forwardKey = key; }

  public void setBackKey(int key){ backKey = key; }

  public void setLeftKey(int key){ leftKey = key; }

  public void setRightKey(int key){ rightKey = key; }

}// end class

